/*****************************************************************
 options.c : processing of options for mps

 Copyright (C) 1999, Richard Urena.  All rights reserved.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.   You can find the GNU
 General Public License in the file COPYING in the source directory,
 or at http://www.gnu.org

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 ****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "options.h"
#include "types.h"
#include "defaults.h"
#include "globals.h"

#define		TempBufferSize		64

/* ------------------------------------------------------------------
   V e r s i o n A s S t r i n g
   ------------------------------------------------------------------ */
char *VersionAsString(void)
{
	static char *vs;
	static int vset = 0;

	if(!vset) {
		if((vs=malloc(128))!=NULL) {
			snprintf(vs,128,"mps v%d.%d.%d  %s",MPS__VERSIONMAJOR,
					MPS__VERSIONMINOR,MPS__PATCHLEVEL,Copyright);
			vs[127] = '\0';
			vset = 1;
		} else
			vs = MPS__EMPTYSTRING;
	}
	return(vs);
}

/* ------------------------------------------------------------------
   O p t i o n V a l u e
   extracts the value of an option (i.e., what comes after
   the "=")
   ------------------------------------------------------------------ */

void OptionValue(char *ValString, char *opt)
{
	int i, j, c, st;

	for(i=0,j=0,st=0; (j<(TempBufferSize-1))&&((c=opt[i])!='\0'); i++) {
		if(!st) {
			if(c=='=')
				st = 1;
		} else
			ValString[j++] = c;
	}
	ValString[j] = '\0';
}

/* ------------------------------------------------------------------
   O n e M a r g i n
   used by several margin options, e.g. -lm=value[units]
   ------------------------------------------------------------------ */
int OneMargin(float *Value, int *Units, char *ValString, char *prog)
{
	double val;
	char *next, temp[TempBufferSize];
	int c, j, ConversionError = 0;

	val = strtod(ValString,&next);
	if(next == ValString) {
		mps_error(MPS__ERRBADMARGINS,prog);
		return(0);
	}
	for(j=0; (j<(TempBufferSize-1))&&((c=(*next))!='\0'); next++)
		temp[j++] = c;
	temp[j] = '\0';
	if(temp[0]!='\0') {
		for(j=0; Measurement[j].Abbrev[0]!='\0' ; j++)
			if(strcmp(temp,Measurement[j].Abbrev)==0)
				break;

		if(Measurement[j].Abbrev[0]=='\0')
			ConversionError = 1;
		else
			(*Units) = j;
	}
	if(ConversionError) {
		mps_error(MPS__ERRBADMARGINS,prog);
		return(0);
	}
	else
		(*Value) = val;

	return(1);
}

/* ==================================================================
   the family of "Opt" functions, process the various mps options
   ================================================================== */

/* ------------------------------------------------------------------
   O p t F o n t
   ------------------------------------------------------------------ */
int OptFont(char *prog, char *ValString)
{
	int i;

	for(i=0; KnownFonts[i].Name[0]!='\0'; i++) {
		if(strcmp(ValString,KnownFonts[i].Name)==0) {
			Font = &(KnownFonts[i]);
			return(1);
		}
	}
	mps_error(MPS__ERRBADFONT,prog);
	fprintf(stderr,"     (Known fonts:");
	for(i=0; KnownFonts[i].Name[0]!='\0'; i++)
		fprintf(stderr," %s",KnownFonts[i].Name);
	fprintf(stderr,")\n");
	return(0);
}

/* ------------------------------------------------------------------
   O p t P a p e r
   ------------------------------------------------------------------ */
int OptPaper(char *prog, char *ValString)
{
	int i;

	for(i=0; KnownPapers[i].Name[0]!='\0'; i++) {
		if(strcmp(ValString,KnownPapers[i].Name)==0) {
			Paper = &(KnownPapers[i]);
			return(1);
		}
	}
	mps_error(MPS__ERRBADPAPER,prog);
	fprintf(stderr,"     (Known papers:");
	for(i=0; KnownPapers[i].Name[0]!='\0'; i++)
		fprintf(stderr," %s",KnownPapers[i].Name);
	fprintf(stderr,")\n");
	return(0);
}

/* ------------------------------------------------------------------
   O p t D u p l e x  ,  O p t N o D u p l e x
   ------------------------------------------------------------------ */
int OptDuplex(char *prog, char *ValString)
{
	AssumeDuplex = 1;
	return(1);
}

int OptNoDuplex(char *prog, char *ValString)
{
	AssumeDuplex = 0;
	return(1);
}

/* ------------------------------------------------------------------
   O p t T r u n c  ,  O p t N o T r u n c
   ------------------------------------------------------------------ */
int OptTrunc(char *prog, char *ValString)
{
	FoldLong = 0;
	return(1);
}

int OptNoTrunc(char *prog, char *ValString)
{
	FoldLong = 1;
	return(1);
}

/* ------------------------------------------------------------------
   O p t P a g e N u m 1
   ------------------------------------------------------------------ */
int OptPageNum1(char *prog, char *ValString)
{
	PageNumbers = 1;
	return(1);
}

/* ------------------------------------------------------------------
   O p t P a g e N u m 2
   (called when -pn is followed by =)
   ------------------------------------------------------------------ */
int OptPageNum2(char *prog, char *ValString)
{
	int pn;
	char *endptr;

	pn = strtol(ValString,&endptr,10);
	if((endptr != ValString) && ((*endptr) == '\0') && (pn > 0)) {
		PageNumbers = 1;
		PageNo = pn;
		return(1);
	}
	mps_error(MPS__ERRBADPAGENO,prog);
	return(0);
}

/* ------------------------------------------------------------------
   O p t N o P a g e N u m
   ------------------------------------------------------------------ */
int OptNoPageNum(char *prog, char *ValString)
{
	PageNumbers = 0;
	return(1);
}

/* ------------------------------------------------------------------
   O p t T a b s
   ------------------------------------------------------------------ */
int OptTabs(char *prog, char *ValString)
{
	int te;
	char *nonum;

	te = strtol(ValString,&nonum,10);
	if( (nonum != ValString) && ((*nonum) == '\0') 
			&& (te >= 1) && (te <= 16) )				{
		TabStop = te;
		return(1);
	}
	mps_error(MPS__ERRBADTABSTOP,prog);
	return(0);
}

/* ------------------------------------------------------------------
   O p t M a r g i n s
   ------------------------------------------------------------------ */
int OptMargins(char *prog, char *ValString)
{
	char *current, *next, temp[TempBufferSize];
	double Value[4], val;
	int i, j, c, ConversionError, Units[4];

	Value[0] = Margin.Left.Measure;
	Value[1] = Margin.Right.Measure;
	Value[2] = Margin.Top.Measure;
	Value[3] = Margin.Bottom.Measure;
	for(i=0; i<4; i++)
		Units[i] = MPS__DEFAULTUNITS;

	for(ConversionError=0,current=ValString,i=0 ;
			(i<4) && ((*current)!='\0') ;  i++,current=next) 	{

		val = strtod(current,&next);

		if((c=(*next))==',') {
			if(next!=current)
				Value[i] = val;
			++next;
			continue;
		} else if(c=='\0') {
			if(next!=current)
				Value[i] = val;
			break;
		} else if(next==current) {
			ConversionError = 1;
			break;
		}

		for(j=0; (j<(TempBufferSize-1))&&((c=(*next))!=',')&&(c!='\0');
									next++)
			temp[j++] = c;
		temp[j] = '\0';
		if(temp[0]!='\0') {
			for(j=0; Measurement[j].Abbrev[0]!='\0' ; j++)
				if(strcmp(temp,Measurement[j].Abbrev)==0)
					break;
	
			if(Measurement[j].Abbrev[0]=='\0')
				ConversionError = 1;
			else
				Units[i] = j;
		}

		if(ConversionError)
			break;
		else
			Value[i] = val;

		/* OK, we appear to be ready to move on to the next one */

		for( ; ((c=(*next))!=',')&&(c!='\0'); next++)
			/* nothing */ ;
		if(c==',')
			++next;

	}

	if((i<3) || ConversionError) {
		mps_error(MPS__ERRBADMARGINS,prog);
		return(0);
	}

	/* at this point, the margins specification is syntactically
	   correct.  whether they are consistent or reasonable
	   is checked later, from the main program
	*/

	Margin.Left.Measure = Value[0];
	Margin.Right.Measure = Value[1];
	Margin.Top.Measure = Value[2];
	Margin.Bottom.Measure = Value[3];

	Margin.Left.Units = Units[0];
	Margin.Right.Units = Units[1];
	Margin.Top.Units = Units[2];
	Margin.Bottom.Units = Units[3];

	return(1);
}

/* ------------------------------------------------------------------
   O p t L e f t M a r g i n  ,  O p t R i g h t M a r g i n
   O p t T o p M a r g i n  ,   O p t B o t t o m M a r g i n
   ------------------------------------------------------------------ */
int OptLeftMargin(char *prog, char *ValString)
{
	return(OneMargin(&Margin.Left.Measure,&Margin.Left.Units,ValString,prog));
}

int OptRightMargin(char *prog, char *ValString)
{
	return(OneMargin(&Margin.Right.Measure,&Margin.Right.Units,ValString,prog));
}

int OptTopMargin(char *prog, char *ValString)
{
	return(OneMargin(&Margin.Top.Measure,&Margin.Top.Units,ValString,prog));
}

int OptBottomMargin(char *prog, char *ValString)
{
	return(OneMargin(&Margin.Bottom.Measure,&Margin.Bottom.Units,ValString,prog));
}

/* ------------------------------------------------------------------
   O p t H e l p
   ------------------------------------------------------------------ */
int OptHelp(char *prog, char *ValString)
{
	char *fmt1 = "   %-20s  %s\n";

	fprintf(stderr,"\nUsage: mps [option...] [file...]\n");
	fprintf(stderr,"\nOptions:\n\n");

	fprintf(stderr,fmt1,"-font=<font>","use <font> (ex: -font=courier9)");
	fprintf(stderr,fmt1,"-paper=<paper>","assume <paper> (ex: -paper=letter)");
	fprintf(stderr,fmt1,"-lm=<size><unit>",
			"set left margin only (ex: -lm=0.5in)");
	fprintf(stderr,fmt1,"-rm=<size><unit>","set right margin only");
	fprintf(stderr,fmt1,"-tm=<size><unit>","set top margin only");
	fprintf(stderr,fmt1,"-bm=<size><unit>","set bottom margin only");
	fprintf(stderr,fmt1,"-margins=<l,r,t,b>",
				"set all margins at once");
	fprintf(stderr,fmt1,"-pn[=num]","print page numbers");
	fprintf(stderr,fmt1,"-nopn","don't print page numbers");
	fprintf(stderr,fmt1,"-trunc, -notrunc","truncate/fold long lines");
	fprintf(stderr,fmt1,"-tabs=<n>","expand tabs to <n> spaces");
	fprintf(stderr,fmt1,"-duplex,-nouplex",
				"assume/don't assume double-sided printing");
	fprintf(stderr,fmt1,"--help","(this screen)");
	fprintf(stderr,fmt1,"--settings","print current settings");
	fprintf(stderr,fmt1,"--version","print version");


	fputc('\n',stderr);
	return(0);
}

/* ------------------------------------------------------------------
   O p t S e t t i n g s
   ------------------------------------------------------------------ */
int OptSettings(char *prog, char *ValString)
{
	char *fmt1 = "\t%-18s :  ";

	fprintf(stderr,"\nCurrent settings:\n\n");
	
	fprintf(stderr,fmt1,"Font ");
	fprintf(stderr,"%s\n",Font->Name);

	fprintf(stderr,fmt1,"Paper ");
	fprintf(stderr,"%s\n",Paper->Name);

	fprintf(stderr,fmt1,"Margins (l,r,t,b)");
	fprintf(stderr,"%4.2f%s, %4.2f%s, %4.2f%s, %4.2f%s\n",
		Margin.Left.Measure,Measurement[Margin.Left.Units].Abbrev,
		Margin.Right.Measure,Measurement[Margin.Right.Units].Abbrev,
		Margin.Top.Measure,Measurement[Margin.Top.Units].Abbrev,
		Margin.Bottom.Measure,Measurement[Margin.Bottom.Units].Abbrev);

	fprintf(stderr,fmt1,"Tabs expand to");
	fprintf(stderr,"%d characters\n",TabStop);

	fprintf(stderr,fmt1,"Assume duplex");
	if(AssumeDuplex)
		fprintf(stderr,"yes");
	else
		fprintf(stderr,"no");
	fputc('\n',stderr);

	
	fprintf(stderr,fmt1,"Long lines ");
	if(FoldLong)
		fprintf(stderr,"folded");
	else
		fprintf(stderr,"truncated");
	fputc('\n',stderr);

	fprintf(stderr,fmt1,"Page numbers ");
	if(PageNumbers)
		fprintf(stderr,"yes");
	else
		fprintf(stderr,"no");
	fputc('\n',stderr);

	if(PageNumbers) {
		fprintf(stderr,fmt1,"First page ");
		fprintf(stderr,"%d\n",PageNo);
	}

	fprintf(stderr,fmt1,"Default units ");
	fprintf(stderr,"%s\n",Measurement[MPS__DEFAULTUNITS].Abbrev);


	fputc('\n',stderr);
	return(0);
}

/* ------------------------------------------------------------------
   O p t V e r s i o n
   ------------------------------------------------------------------ */
int OptVersion(char *prog, char *ValString)
{
	fprintf(stderr,"%s\n",VersionAsString());
	return(0);
}

/* ------------------------------------------------------------------
   O p t U n k n o w n
   ------------------------------------------------------------------ */
int OptUnknown(char *prog, char *ValString)
{
	mps_error(MPS__ERRBADOPTION,prog);
	return(0);
}

/* ===================================================================
   action table for mps options
   =================================================================== */

struct t_optinfo {
	char *String;						/* string representation */
	int   FullMatch;					/* full or partial match? */
	int (*Action)(char *, char *);		/* action for this option */

} OptInfo[] = {
	{ "NODASH!!", 1, OptUnknown }, /* this must be the first one */
	{ "-font=", 0, OptFont },
	{ "-paper=", 0, OptPaper },
	{ "-duplex", 1, OptDuplex },
	{ "-noduplex", 1, OptNoDuplex },
	{ "-trunc", 1, OptTrunc },
	{ "-notrunc", 1, OptNoTrunc },
	{ "-pn", 1, OptPageNum1 },
	{ "-pn=", 0, OptPageNum2 },
	{ "-nopn", 1, OptNoPageNum },
	{ "-tabs=", 0, OptTabs },
	{ "-margins=", 0, OptMargins },
	{ "-lm=", 0, OptLeftMargin},
	{ "-rm=", 0, OptRightMargin},
	{ "-tm=", 0, OptTopMargin},
	{ "-bm=", 0, OptBottomMargin},
	{ "--help", 1, OptHelp},
	{ "--settings", 1, OptSettings},
	{ "--version", 1, OptVersion },
	{ MPS__EMPTYSTRING, 1, NULL }	/* this must be the last one */
};

/* ------------------------------------------------------------------
   P a r s e O p t i o n
   ------------------------------------------------------------------ */
int ParseOption(char *optn, char *prog)
{
	int i, match;
	char ValString[TempBufferSize];

	if(optn[0] == MPS__OPTIONPREFIX) {				/* safety */
		for(i=1; OptInfo[i].String[0] != '\0'; i++) {
			if(OptInfo[i].FullMatch)
				match = (strcmp(optn,OptInfo[i].String) == 0);
			else
				match = (strstr(optn,OptInfo[i].String) == optn);
			if(match) {
				OptionValue(ValString,optn);
				return(OptInfo[i].Action(prog,ValString));
			}
		}
	}
	return(OptInfo[0].Action(prog,MPS__EMPTYSTRING));	/* bad option */
}

#undef		TempBufferSize

